/*
 * The contents of this file are copyright 2012 Clinton Bale
 * <clinton.bale@gmail.com>
 *
 * This work is licensed under the Creative Commons "Attribution-Share Alike 3.0 Unported" License.
 * To view a copy of this license, visit http://creativecommons.org/licenses/by-sa/3.0/legalcode
 * or, send a letter to Creative Commons, 171 2nd Street, Suite 300, San Francisco, California, 94105, USA.
*/

#include "mazegenerator.h"
#include "mapfile.h"
#include "timer.h"

#include <cstdio>
#include <cstdlib>
#include <windows.h>

static size_t	g_mapSizeX = 20;
static size_t	g_mapSizeY = 20;
static size_t	g_mapHeight = 192;
static char		g_mapOutput[MAX_PATH+1] = "";
static char		g_wallTexture[32] = "C2A4E_W1C";
static char		g_floorTexture[32] = "C1A3FLOOR02";
static bool		g_nullOptimize = true;
static bool		g_showPreview = false;
static bool	    g_verbose = false;

#define WALLSIZE 64

int index_of(const char* str, char ch) {
	int index = 0;
	if(*str) {
		do {
			if(*str == ch)
				return index;
			++index;
		} while(*str++);
	}
	return -1;
}

bool file_exists(const char* path) {
	FILE* fp = fopen(path, "r");
	bool flag = false;
	if(fp) {
		flag = true;
		fclose(fp);
	}
	return flag;
}

int main(int argc, char* argv[]) {	
	char* argstr;
	int arglen, mid;
	try {
		if(argc < 2)
			throw "";
		for(int i = 1; i < argc; ++i) {
			argstr = argv[i];			
			if(argstr[0] == '-') {
				arglen = strlen(argstr);
				switch(argstr[1]) 
				{
				case 's':
					if(i+1 > argc-1)
						throw "-s found without dimensions (width,height)";
					mid = index_of(argv[i+1], ',');
					if(mid >= 0) {
						argv[i+1][mid] = 0;
						
						g_mapSizeX = atoi(argv[i+1]);
						g_mapSizeY = atoi((char*)(argv[i+1]+mid+1));
					}
					else 
					{
						g_mapSizeY = g_mapSizeX = atoi(argv[i+1]);
					}
					if(g_mapSizeX <= 2 || g_mapSizeY <= 2 || g_mapSizeX * g_mapSizeY <= 10)
						throw "Total maze area must be greater than 10.";
					if(g_mapSizeX > 62 || g_mapSizeY > 62)
						throw "Max map dimensions are 62,62";
					i++;
					break;
				case 'm':
					if(argstr[2] == 'h') {
						if(i+1 > argc-1)
							throw "-mh found without a number.";
						g_mapHeight = atoi(argv[i+1]);
						if(g_mapHeight < 128)
							throw "Maze height must be greater than 127";
						i++;
					}
					break;
				case 't':
					if(argstr[2] == 'w')
					{
						if(i+1 > argc-1)
							throw "-tw found without a texture name.";
						if(strlen(argv[i+1]) >= sizeof(g_wallTexture))
							throw "Texture name too long";
						strcpy(g_wallTexture, argv[i+1]);
						if(strlen(g_wallTexture) <= 0)
							throw "Invalid texture name";
						i++;
					}
					else if(argstr[2] == 'f')
					{
						if(i+1 > argc-1)
							throw "-tf found without a texture name.";
						if(strlen(argv[i+1]) >= sizeof(g_floorTexture))
							throw "Texture name too long";
						strcpy(g_floorTexture, argv[i+1]);
						if(strlen(g_floorTexture) <= 0)
							throw "Invalid texture name";
						i++;
					}					
					break;
				case 'n':
					if(argstr[2] == 't')
						g_nullOptimize = false;
					break;
				case 'p':
					g_showPreview = true;
					break;
				case 'v':
					g_verbose = true;
					break;
				case 'h':
					throw "";
					break;		
				}				
			}
			else {
				if(!*g_mapOutput) {
					strcpy(g_mapOutput, argstr);
					if(g_verbose && file_exists(g_mapOutput)) {
						printf("Warning: overwriting existing .map file\n");
					}
				}
			}
		}
		if(!*g_mapOutput)
			throw "No output file specified";
	}
	catch(const char* ex) {
//		Win32 console size
//	       0         1         2         3         4         5         6         7         8
//	        12345678901234567890123456789012345678901234567890123456789012345678901234567890
		if(*ex)
			printf("Command line error: %s\n", ex);
		printf(
			"mazegen v1.0 rel\n"
			"Submit detailed bug reports to (clinton.bale@gmail.com)\n\n"
			"  Usage: mazegen <outfile>[.map] [options]\n"
			"    -h                  Displays this help screen\n"
			"    -v                  Show extended information\n"
			"    -p                  Show a preview of a the maze in the console\n"
			"    -s <width,height>   The size of the generated map (default is 20,20)\n"
			"    -mh <number>        Specifies the height of the maze (default is 192)\n"
			"    -tw <texname>       Specifies the wall texture to be used\n"
			"                        (default is C2A4E_W1C)\n"
			"    -tf <texname>       Specifies the floor texture to be used\n"
		    "                        (default is C1A3FLOOR02)\n"
			"    -nt                 Don't optimize brush face textures with NULL\n");
 		return 1;
	}

	Timer timer;
	timer.start();	
	
	printf("Generating maze... ");
	if(g_verbose) {
		printf("\nUsing maze dimensions w: %d, h: %d\n", g_mapSizeX, g_mapSizeY);
	}

	MazeGenerator gen(g_mapSizeX, g_mapSizeY);
	gen.generate(0, 0);

	int mazeWidth, mazeHeight;
	char** mazeData = gen.print(false, &mazeWidth, &mazeHeight);

	if(g_showPreview) {
		printf("\n");
		gen.print();
	}

	printf("done\n");

	printf("Generating .map file... ");

	if(g_verbose) {
		printf("\nTotal number of cells: %d\n", mazeWidth*mazeHeight);
		printf("Using map height %d\n", g_mapHeight);
		printf("Using wall texture '%s'\n", g_wallTexture);
		printf("Using floor texture '%s'\n", g_floorTexture);		
		if(!g_nullOptimize) {
			printf("Warning: not optimizing brush faces with NULL texture\n");
		}
	}

	MapFile map(g_mapOutput);
	if(!map.create()) {
		printf("Error creating .map file\n");
		return 1;
	}

	for(int y = 0; y < mazeHeight; y++) {
		for(int x = 0; x < mazeWidth; x++) {
			if(mazeData[y][x] == '#') {
				map.create_brush((x-(mazeWidth/2))*WALLSIZE, (y-(mazeHeight/2))*WALLSIZE, 0, WALLSIZE, WALLSIZE, g_mapHeight, g_wallTexture, g_nullOptimize ? TOP | BOTTOM : 0);
			}
		}
	}
	map.create_brush(0-(mazeWidth/2)*WALLSIZE, 0-(mazeHeight/2)*WALLSIZE, g_mapHeight, mazeHeight*WALLSIZE, mazeWidth*WALLSIZE, 16, "SKY", 0);
	map.create_brush(0-(mazeWidth/2)*WALLSIZE, 0-(mazeHeight/2)*WALLSIZE, -16, mazeHeight*WALLSIZE, mazeWidth*WALLSIZE, 16, g_floorTexture, g_nullOptimize ? ~(TOP) : 0);

	if(!map.close()) {
		printf("Error closing .map file\n");
		return 1;
	}
	
	printf("done\n");
	timer.stop();
	printf("%lf seconds elapsed\n", timer.getElapsedTimeInSec());

	return 0;
}