/*
 * The contents of this file are copyright 2012 Clinton Bale
 * <clinton.bale@gmail.com>
 *
 * This work is licensed under the Creative Commons "Attribution-Share Alike 3.0 Unported" License.
 * To view a copy of this license, visit http://creativecommons.org/licenses/by-sa/3.0/legalcode
 * or, send a letter to Creative Commons, 171 2nd Street, Suite 300, San Francisco, California, 94105, USA.
*/

#include "mazegenerator.h"

#include <cstdio>
#include <ctime>
#include <cassert>

#define WALLCHAR '#'
#define PATHCHAR ' '

MazeGenerator::MazeGenerator(size_t w, size_t h)
	: width(w), height(h)
{
	assert(width >= 5);
	assert(height >= 5);
	assert(width * height >= 25);

	maze = new cell_t*[height];
	for(size_t i = 0; i < height; i++) {
		maze[i] = new cell_t[width];
	}
	this->initialize();

	srand((size_t)time(NULL));
}

void MazeGenerator::initialize(void)
{
	for(size_t y = 0; y < height; y++) {
		for(size_t x = 0; x < width; x++)
		{
			maze[y][x].walls = ALL;
			maze[y][x].y = y;
			maze[y][x].x = x;
		}
	}
}

bool MazeGenerator::get_neighbors(int xpos, int ypos, std::vector<cell_t*>& out)
{
	out.clear();
	for (int x = -1; x <= 1; x++) {
		for (int y = -1; y <= 1; y++)
		{
			if(xpos + x < (int)width && ypos + y < (int)height && xpos + x >= 0 &&
				ypos + y >= 0 && (x == 0 || y == 0) && x != y)
			{
				if(maze[ypos+y][xpos+x].walls == 0xF)
					out.push_back(&maze[ypos+y][xpos+x]);
			}
		}
	}
	return out.size() > 0;
}

void MazeGenerator::generate(void) {
	generate(width/2, height/2);
}

void MazeGenerator::generate(int startX, int startY)
{
	assert(startX < (int)width || startX >= 0);
	assert(startY < (int)height || startY >= 0);

	this->initialize();

	cell_t* currentCell = &maze[startY][startX];
	cell_t* newCell = (cell_t*)0;

	unsigned int visitedCells = 1;

	std::vector<cell_t*> neighbors;
	std::stack<cell_t*> cellStack;

	neighbors.reserve(4);
	while(visitedCells < this->total_cells()) {
		if(this->get_neighbors(currentCell->x, currentCell->y, neighbors))
		{
			newCell = neighbors[rand()%neighbors.size()];
			currentCell->remove_wall(newCell);

			cellStack.push(currentCell);
			currentCell = newCell;

			++visitedCells;
		}
		else
		{
			currentCell = cellStack.top();
			cellStack.pop();
		}
	}
}

MazeGenerator::~MazeGenerator(void)
{
	for(size_t i = 0; i < height; ++i) {
		if(maze[i]) 
		{
			delete [] maze[i];
			maze[i] = (cell_t*)0;
		}
	}

	delete [] maze;
	maze = (cell_t**)0;

	while(garbage.size()) {
		if(garbage.top()) {
			delete [] garbage.top();
			garbage.top() = (char*)0;
		}
		garbage.pop();
	}
}

void MazeGenerator::print() {
	this->print(true, NULL, NULL);
}

char** MazeGenerator::print(bool console, int* out_width, int* out_height)
{
	char** output = new char*[height*2+1];	

	if(out_width)
		*out_width = 1+width*2;
	if(out_height)
		*out_height = 1+height*2;

	for(size_t y = 0; y < height*2+1; ++y) {
		output[y] = new char[width*2+2];
		output[y][width*2] = WALLCHAR;
		output[y][width*2+1] = '\0';
		for(size_t x = 0; x < width*2; x++)
		{
			if(x % 2 == 0 || y % 2 == 0)			 
				output[y][x] = WALLCHAR;			 
			else
				output[y][x] = PATHCHAR;
		}
	}

	for(size_t y = 1, yy = 0; y < height*2; y += 2, ++yy) {
		for(size_t x = 1, xx = 0; x < width*2; x += 2, ++xx) {
			if(!(maze[yy][xx].walls & SOUTH))
				output[y+1][x] = ' ';
			if(!(maze[yy][xx].walls & NORTH))
				output[y-1][x] = ' ';
			if(!(maze[yy][xx].walls & WEST))
				output[y][x-1] = ' ';
			if(!(maze[yy][xx].walls & EAST))
				output[y][x+1] = ' ';
		}
	}

	for(size_t y = 0; y < height*2+1; ++y) {		
		if(console) {
			printf("%s\n", output[y]);
			delete [] output[y];
		}
		else
			garbage.push(output[y]);
	}
	if(console) {
		delete [] output;
		return (char**)0;
	}
	else {
		garbage.push(output);
		return output;
	}	
}
